/*-------------------------------------------------------------*/
/*This script is written for MENG Project: Heart Valve Bioreactor
  by Kang Li (kl694), Cornell University*/
#define F_CPU 16000000UL
#include "trtSettings.h"
#include "trtkernel_1284.c"
#include <avr/interrupt.h>
#include <avr/pgmspace.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <avr/sleep.h>

#define begin {
#define end }

// serial communication library
// Don't mess with the semaphores
#define SEM_RX_ISR_SIGNAL 1
#define SEM_STRING_DONE 2 // user hit <enter>
#include "trtUart.h"
#include "trtUart.c"

#define DIFFERENCE 11
#define AORTA 12
#define VENTRICLE 13

#define READ(U, N) ((U) >> (N) & 1u)
#define SET(U, N) ((void)((U) |= 1u << (N)))
#define CLR(U, N) ((void)((U) &= ~(1u << (N))))
#define FLIP(U, N) ((void)((U) ^= 1u << (N)))

#define SEM_SEND 3
#define SEM_INPUT 4
#define SEM_SQUARE 5
 
#define T_OVERFLOW 500
// UART file descriptor
// putchar and getchar are in uart.c
FILE uart_str = FDEV_SETUP_STREAM(uart_putchar, uart_getchar, _FDEV_SETUP_RW);

int args[2];
unsigned char  P_tank_H, P_tank_L;
double P_diff;
int count;
double deltaP;
double delta;
double time_squarewave, dP_MA, Gain, Bias;
int thres,T;
float P, C, D, G, B;
unsigned char P_aorta, P_ventricle;
unsigned char channel_flag ;
long HighTank_total, LowTank_total;
int pumpSignal;
char sendData[14];


void initialize(void);

int main(void)
begin
	initialize();
	set_sleep_mode(SLEEP_MODE_IDLE);
 	sleep_enable();
 	while (1) 
  	begin
		CLR(PORTD,1);
  		sleep_cpu();
	end
end

void pump_control(void * args)
begin
	uint32_t rel,dead;
	while(1)
	begin
		switch(channel_flag)
		begin
			case (DIFFERENCE):				
				
				// wait for ADC done
				while (ADCSRA & (1<<ADSC));
				P_tank_H = ADCH;
				ADMUX = (1<<REFS0) | (1<<ADLAR) + 0b00001; 
				ADCSRA |= (1<<ADSC) ;
				HighTank_total += P_tank_H; 	
				// wait for ADC done
				while (ADCSRA & (1<<ADSC));
				P_tank_L = ADCH;
				//Reference voltage is 5V; Write one to ADLAR to left adjust the result; Select the PINA1(ADC1) as analog input
				ADMUX = (1<<REFS0) | (1<<ADLAR) + 0b00010;
				ADCSRA |= (1<<ADSC) ;
				LowTank_total += P_tank_L;				
				count++;
				if (count >= 50)
				begin
					count = 0;
					
					trtWait(SEM_INPUT); 					
					deltaP = ((float)HighTank_total - (float)LowTank_total) / 50 * Gain;
					trtSignal(SEM_INPUT);
					HighTank_total = 0;
					LowTank_total = 0;
					
					//load the values sent from MATLAB and protect the variables using TRT
					trtWait(SEM_INPUT);
					P_diff = P;
					dP_MA = C;
					delta = D;
					Gain = G;
					Bias = B;
					thres = T;
					trtSignal(SEM_INPUT);
					
					//if P_diff, dP_MA, delta are all zeros, don't start the pump
					if((P_diff > 0) || (dP_MA > 0) || (delta > 0))
					begin
						if(deltaP < P_diff - delta * dP_MA)
						begin
							SET(PORTD, 2); // turn on the pump
							trtWait(SEM_SEND);
							pumpSignal = 1;
							trtSignal(SEM_SEND);
						end
						if(deltaP >= P_diff + delta * dP_MA)
						begin
							CLR(PORTD, 2); // turn off the pump
							trtWait(SEM_SEND);
							pumpSignal = 0;
							trtSignal(SEM_SEND);
						end					
					end
					else
					begin
						CLR(PORTD, 2);
					end				
				end		
				 
				channel_flag = VENTRICLE;
				break;
			case (VENTRICLE):				
				
				// wait for ADC done
				while (ADCSRA & (1<<ADSC));
				P_ventricle = ADCH;
				//Reference voltage is 5V; Write one to ADLAR to left adjust the result; Select the PINA3(ADC3) as analog input
				ADMUX = (1<<REFS0) | (1<<ADLAR) + 0b00011;
				ADCSRA |= (1<<ADSC) ;
				channel_flag = AORTA;
				break;
			case (AORTA):
				
				while (ADCSRA & (1<<ADSC));
				P_aorta = ADCH;;	

				if(P_aorta >= thres)
				begin
					CLR(PORTC,0); // open the two-way solenoid valve
				end
				else 
				begin
					SET(PORTC,0); // close the two-way solenoid vale (the valve is closed when energized)
				end
				
				//Reference voltage is 5V; Write one to ADLAR to left adjust the result; Select the PINA0(ADC0) as analog input
				ADMUX = (1<<REFS0) | (1<<ADLAR) + 0b00000; 
				ADCSRA |= (1<<ADSC) ;

				trtWait(SEM_SEND);
				// put the four variables in a package
				sprintf(sendData, "H%3dL%3dV%3dA%3dU%1d",P_tank_H, P_tank_L, P_ventricle, P_aorta, pumpSignal);
				trtSignal(SEM_SEND);
				channel_flag = DIFFERENCE;
				break;
		
		end	


		rel = trtCurrentTime() + SECONDS2TICKS(0.01);
		dead = trtCurrentTime() + SECONDS2TICKS(0.011);
		trtSleepUntil(rel, dead);

	end
end

void PC_communication(void * args)
begin
	char cmd[40], P_str[10], C_str[10], D_str[10], G_str[10], B_str[10], T_str[10];
	int p, c, d, g, b, t;
	
	while(1)
	begin	
		fscanf(stdin, "%s", cmd) ;
		
		if (cmd[0] == 'o') // change the value of the protected variables
		begin
			p = strchr(cmd,'P')-cmd;
			c = strchr(cmd,'C')-cmd;
			d = strchr(cmd,'D')-cmd;
			g = strchr(cmd,'G')-cmd;
			b = strchr(cmd,'B')-cmd;
			t = strchr(cmd,'T')-cmd;
			strlcpy(P_str, cmd + p + 1, c - p);
			strlcpy(C_str, cmd + c + 1, d - c);
			strlcpy(D_str, cmd + d + 1, g - d);
			strlcpy(G_str, cmd + g + 1, b - g);
			strlcpy(B_str, cmd + b + 1, t - b);
			strlcpy(T_str, cmd + t + 1, strlen(cmd) - t);
			trtWait(SEM_INPUT);		
			sscanf(P_str, "%f", &P);
			sscanf(C_str, "%f", &C);
			sscanf(D_str, "%f", &D);
			sscanf(G_str, "%f", &G);
			sscanf(B_str, "%f", &B);
			sscanf(T_str, "%d", &T);									
			trtSignal(SEM_INPUT);
		end
		
		if (cmd[0] == 's') // send the package to PC
		begin
			trtWait(SEM_SEND);
			fprintf(stdout, "%s", sendData);
			trtSignal(SEM_SEND) ;
		end
		
		if (cmd[0] == 'c') // check the current state of the whole system
		begin
			trtWait(SEM_INPUT);
 			fprintf(stdout, "\rP_diff%fdP_MA%fdelta%fGain%fBias%fThres%d\n", P_diff, dP_MA, delta, Gain, Bias, thres);
			trtSignal(SEM_INPUT);
		end
	end
end

void initialize(void)
begin
	CLKPR = (1 << CLKPCE); 
  	CLKPR = (1 << CLKPS0) | (1 << CLKPS1); // set divide by 8

	//PINA is input
	DDRA = 0x00;
	PORTA = 0x00;
	//PIND is output
	DDRD = 0xff;
	PORTD = 0x00;
	//PINC is output
	DDRC = 0xff;
	PORTC = 0x00;
	
	ADMUX = (1<<REFS0) | (1<<ADLAR) + 0b00000;
	ADCSRA = (1<<ADEN) | (1<<ADSC) + 7 ;	
	
	channel_flag = DIFFERENCE;
	P_tank_H = 0;
	P_tank_L = 0;
	P_aorta = 0;
	P_ventricle = 0;
	P_diff = 0;
	delta = 0;
	dP_MA = 0;
	Gain = 0;
	Bias = 0;
	time_squarewave = 0;
	thres = 0;
	count = 0;

	T=0;
	
	//ini uart
	trt_uart_init();
	stdout = stdin = stderr = &uart_str;	
//	fprintf(stdout,"\n\r...start...\r");
	
	trtInitKernel(300);	
	trtCreateSemaphore(SEM_RX_ISR_SIGNAL, 0) ; // uart receive ISR semaphore
	trtCreateSemaphore(SEM_STRING_DONE,0) ;  // user typed <enter>
	trtCreateSemaphore(SEM_SEND, 1) ; 
	trtCreateSemaphore(SEM_INPUT, 1) ;
	trtCreateSemaphore(SEM_SQUARE, 1) ;
		
	trtCreateTask(PC_communication, 300, SECONDS2TICKS(0.15), SECONDS2TICKS(0.25), &(args[0]));
	trtCreateTask(pump_control, 300, SECONDS2TICKS(0.01), SECONDS2TICKS(0.015), &(args[1]));
	sei();
end
